#ifndef ACONFIG_H_
#define ACONFIG_H_

#include <common.h>

#define ETH_KEEP        1
#define ETH_DROP        0

#define ACONFIG_MAGIC   (0x41434F4E)            //long integer which represent "ACON".
                                        //Used to differentiate the protocol of the packet.
#define ACONFIG_REVISION        1               //Version header - for major change in packet format
#define ACONFIG_BUILD           1               //Version build

#define ACONFIG_SERVER_PORT_NUM         5070    /* server's port number for bind() */
#define ACONFIG_REQUEST_MSG_SIZE        1060    /* max size of request message */

#define ACONFIG_PROTOSPACE      0xAAAA  /* temporarily */
#define ACONFIG_HDR_LEN         48      /* aconfig header length */
#define ACONFIG_DHDR_LEN        8       /* aconfig data header length exclude data*/

#define ACONFIG_VERSION       "1.0.6"

//#define ACONFIG_DEBUG		1

#if defined(ACONFIG_DEBUG)
#define ACONFIG_PRINTF(s, ARGS...) printf(s, ##ARGS)
#else
#define ACONFIG_PRINTF(s, ARGS...)
#endif

#define ENET_ADDR_LEN 		6

//Aconfig packet header
//Total bytes : 48
typedef struct
{
	unsigned long   magic;			/* magic num for aconfig identification */
	unsigned char   revision;		/* revision, 1 for IPv4 & 2 for IPv6, only 1 supported right now */
	unsigned char   build[3];		/* build revision */
	short           encryption_mode;        /* encryption mode, 0 for none, other for encryptions */
	short           id;                     /* packet identifier */
	short           cookie;                 /* cookie, not used */
	short           operation;              /* operation */
	unsigned long   src_ip;                 /* source IP addr */
	unsigned long   dst_ip;                 /* dest IP addr */
	short           pkt_seq;                /* packet sequence */
	short           last_pkt;               /* last packet flag, 0 for no, 1 for yes */
	char            reserved[20];           /* reserved */
} aconfigHeader;

/* aconfig data header */
typedef struct
{
	unsigned long   type;
	short           coding;
	short           length;
} aconfigDataHeader;

typedef struct
{
	aconfigDataHeader data_hdr;
	uchar * pdata;
} aconfigData;

typedef enum
{
	TYPE_END = 0,
	TYPE_DEVICEID,
	TYPE_IPADDR,
	TYPE_SUBNET,
	TYPE_MACADDR,
	TYPE_URL,
	TYPE_CMDLOGIN,
	TYPE_CMDLOGOUT,
	TYPE_FWUPPASSIVE,
	TYPE_FWUPREQUEST,
	TYPE_FWUPREQACK, /* 10 */
	TYPE_FWUPREADY,
	TYPE_FWUPDATA,
	TYPE_FWUPDATAACK,
	TYPE_FWUPFLASHING,
	TYPE_FWUPCOMPLETED,
	TYPE_FWUPFAILED,
	TYPE_FWUPSIZE,
	TYPE_FWUPSTATUS,
	TYPE_CMDRESETPSI = 42,
	TYPE_PSISIZE = 50,
	TYPE_BOARDID,
	TYPE_NUMMACADDR,
	TYPE_PRODUCTID,
	TYPE_CUSTOMERID,
	TYPE_AZTECHENV,
	TYPE_UBOOTACONFIGVERSION,
} aconfigType;

typedef enum
{
	OP_GET = 0,
	OP_SET,
	OP_REPLY,
	OP_AUTH,
	OP_FWUP, /* upgrade firmware */
	OP_CMD,
} acconfigOperation;

typedef enum
{
	CODING_STRING = 0,
	CODING_HEX,
	CODING_ACK,
	CODING_NAK,
	CODING_INVALID,
} aconfigCoding;

typedef enum
{
	CMD_STATUS_FAILED = 0,
	CMD_STATUS_SUCCESS,
	CMD_STATUS_DATA_SAME,                   /* data is same with that in NVRAM */
	CMD_STATUS_DATA_INVALID,                /* data is invalid */
	CMD_STATUS_DATA_LENGTH_INVALID,         /* data length is invalid, too large, too small, etc */
	CMD_STATUS_BOARDID_INVALID,             /* Board ID is invalid */
} aconfigCmdStatus;

/* structure of aconfig information */
typedef struct
{
	uchar dest_hwaddr[ENET_ADDR_LEN];
	/* FIXME: Check if needed */
	// ether_info_t * eth_info;
	aconfigHeader * hdr;
	aconfigData * data;
} aconfigInfo;

/* statuses we may meet when upgrading */
typedef enum {
	STATUS_VOID = 0,
	STATUS_CONNECTING,
	STATUS_READY,
	STATUS_DATAING,
	STATUS_FLASHING,
	STATUS_COMPLETED,
	STATUS_FAILED,
	STATUS_TIMEOUT,
} aconfigStatus;

typedef enum {
	ERROR_IMAGE_CHECK = 1,          /* image check error */
	ERROR_IMAGE_FLASH,              /* image flash error */
	ERROR_HDR_ID,                   /* packet header id error */
	ERROR_HDR_SEQ,                  /* packet header sequence error */
	ERROR_DATA_CODING,              /* data coding error */
	ERROR_DATA_LENGTH,              /* data length error */
	ERROR_IMAGE_TAG,                /* Broadcom image tag error */
	ERROR_IMAGE_SIZE,               /* image size error, too large, not same with that gotten, etc */
	ERROR_IMAGE_CRC,                /* Broadcom image CRC check error */
	ERROR_IMAGE_UBOOT_SIZE,         /* Uboot size error, too large */
	ERROR_IMAGE_AZTECH_TAG,         /* Aztech tag check failed */
	ERROR_IMAGE_FW_VERSION,         /* Firmware Version check failed */
	ERROR_IMAGE_PRODUCT_ID,         /* Product ID check failed */
	ERROR_IMAGE_BOARD_ID,           /* Board ID check failed */
	ERROR_IMAGE_AZTECH_TAG_CRC,     /* Aztech tag CRC check failed */
	ERROR_IMAGE_DECRYPT,            /* Image decryption failed */
} aconfigErrorCode;

typedef struct
{
	volatile uint64_t timer;
	volatile int count;
	volatile int interval;
} aconfigTimer;



#define AZTECH_FIRMWARE_TAG_STRING_LEN          512
#define AZTECH_FIRMWARE_TAG_ENCRYPT_STRING_LEN  128

#define AZTECH_FIRMWARE_TAG_MAGIC       0x415A5447  //Hex representation of  word "AZTG"
#define AZTECH_FIRMWARE_TAG_VERSION     1

typedef enum
{
	TAG_STRING_END = 0,
	TAG_STRING_ERNCRYPT_STRING,
	TAG_STRING_FIRMWARE_VERSION,
	TAG_STRING_BOARD_ID,
} aztechFirmwareTagType;

typedef enum
{
	ENC_NONE = 0,
	ENC_XOR
} aztechFirmwareEncryptMode;

typedef struct
{
	unsigned int            magic;                  /* magic number */
	unsigned short  version;                        /* version */
	unsigned short  encrypt_mode;   /* encrption mode, 0 plaintext, 1 XOR */
	unsigned char   string[AZTECH_FIRMWARE_TAG_STRING_LEN];         /* customer name, firmware version, product id.
								All values will be stored in string in Type, Length, Value format */
	unsigned int            crc_firmware;   /* firmware CRC */
	unsigned int            crc_trailer;            /* trailer CRC */
} aztechFirmwareTag;

typedef struct {
	uint8_t *eb_data;
	unsigned int eb_length;
	uint8_t *dst_addr;
	uint8_t *src_addr;
} ebuf_t;

void aconfigTimerMonitor(void * arg);

void aconfigInitialization(void);
void aconfigUninitialization(void);

int aconfigBroadcastRequest(void);
int aconfigOnReceive(ebuf_t *buf);

#endif// ACONFIG_H_
