/* $Id: minissdp.c,v 1.3 2007/03/01 22:49:11 nanard Exp $ */
/* MiniUPnP project
 * http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
 * (c) 2006 Thomas Bernard
 * This software is subject to the conditions detailed
 * in the LICENCE file provided within the distribution */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include "config.h"
#include "upnpdescstrings.h"
#include "miniupnpdpath.h"
#include "upnphttp.h"
#include "upnpglobalvars.h"
#include "minissdp.h"
#include "IFX_TR64.h"

/* SSDP ip/port */
#define SSDP_PORT (1900)
#define SSDP_MCAST_ADDR ("239.255.255.250")

static int
AddMulticastMembership(int s, const char *ifaddr)
{
    struct ip_mreq imr;         /* Ip multicast membership */

    /* setting up imr structure */
    imr.imr_multiaddr.s_addr = inet_addr(SSDP_MCAST_ADDR);
    /* imr.imr_interface.s_addr = htonl(INADDR_ANY); */
    imr.imr_interface.s_addr = inet_addr(ifaddr);

    if(setsockopt
       (s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void *)&imr,
        sizeof(struct ip_mreq)) < 0) {
        SYSLOG(LOG_ERR, "setsockopt(udp, IP_ADD_MEMBERSHIP): %m");
        return -1;
    }

    return 0;
}

int
OpenAndConfSSDPReceiveSocket(const char *ifaddr,
                             int n_add_listen_addr,
                             const char **add_listen_addr)
{
    int s;
    struct sockaddr_in sockname;

    if((s = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
        SYSLOG(LOG_ERR, "socket(udp): %m");
        return -1;
    }

    memset(&sockname, 0, sizeof(struct sockaddr_in));
    sockname.sin_family = AF_INET;
    sockname.sin_port = htons(SSDP_PORT);
    /* NOTE : it seems it doesnt work when binding on the specific address */
    /* sockname.sin_addr.s_addr = inet_addr(UPNP_MCAST_ADDR); */
    sockname.sin_addr.s_addr = htonl(INADDR_ANY);
    /* sockname.sin_addr.s_addr = inet_addr(ifaddr); */

    if(bind(s, (struct sockaddr *)&sockname, sizeof(struct sockaddr_in)) < 0) {
        SYSLOG(LOG_ERR, "bind(udp): %m");
        close(s);
        return -1;
    }

    if(AddMulticastMembership(s, ifaddr) < 0) {
        close(s);
        return -1;
    }
    while(n_add_listen_addr > 0) {
        n_add_listen_addr--;
        if(AddMulticastMembership(s, add_listen_addr[n_add_listen_addr]) < 0) {
            SYSLOG(LOG_WARNING,
                   "Failed to add membership for address %s. EXITING",
                   add_listen_addr[n_add_listen_addr]);
        }
    }

    return s;
}

/* open the UDP socket used to send SSDP notifications to
 * the multicast group reserved for them */
int
OpenAndConfSSDPNotifySocket(const char *addr)
{
    int s;
    unsigned char loopchar = 0;
    struct in_addr mc_if;
    struct sockaddr_in sockname;

    if((s = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
        SYSLOG(LOG_ERR, "socket(udp_notify): %m");
        return -1;
    }

    mc_if.s_addr = inet_addr(addr);

    if(setsockopt
       (s, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&loopchar,
        sizeof(loopchar)) < 0) {
        SYSLOG(LOG_ERR, "setsockopt(udp_notify, IP_MULTICAST_LOOP): %m");
        close(s);
        return -1;
    }

    if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_IF, (char *)&mc_if, sizeof(mc_if))
       < 0) {
        SYSLOG(LOG_ERR, "setsockopt(udp_notify, IP_MULTICAST_IF): %m");
        close(s);
        return -1;
    }

    memset(&sockname, 0, sizeof(struct sockaddr_in));
    sockname.sin_family = AF_INET;
    sockname.sin_addr.s_addr = inet_addr(addr);

    if(bind(s, (struct sockaddr *)&sockname, sizeof(struct sockaddr_in)) < 0) {
        SYSLOG(LOG_ERR, "bind(udp_notify): %m");
        close(s);
        return -1;
    }

    return s;
}


/*
 * response from a LiveBox (Wanadoo)
HTTP/1.1 200 OK
CACHE-CONTROL: max-age=1800
DATE: Thu, 01 Jan 1970 04:03:23 GMT
EXT:
LOCATION: http://192.168.0.1:49152/gatedesc.xml
SERVER: Linux/2.4.17, UPnP/1.0, Intel SDK for UPnP devices /1.2
ST: upnp:rootdevice
USN: uuid:75802409-bccb-40e7-8e6c-fa095ecce13e::upnp:rootdevice

 * response from a Linksys 802.11b :
HTTP/1.1 200 OK
Cache-Control:max-age=120
Location:http://192.168.5.1:5678/rootDesc.xml
Server:NT/5.0 UPnP/1.0
ST:upnp:rootdevice
USN:uuid:upnp-InternetGatewayDevice-1_0-0090a2777777::upnp:rootdevice
EXT:
 */

/* not really an SSDP "announce" as it is the response
 * to a SSDP "M-SEARCH" */
static void
SendSSDPAnnounce2(int s, struct sockaddr_in sockname,
                  const char *st, int st_len,
                  const char *host, unsigned short port, const char *uuid, int iUuidFlag)
{
    int l, n;
    char buf[512];
    /* TODO : follow guideline from document "UPnP Device Architecture 1.0" put
       in uppercase. DATE: is recommended SERVER: OS/ver UPnP/1.0 miniupnpd/1.0 */
    l = snprintf(buf, sizeof(buf), "HTTP/1.1 200 OK\r\n"
                 "Server: " MINIUPNPD_SERVER_STRING "\r\n"
                 "EXT:\r\n"
#ifdef IFX_DEFINE_TR64
                 "Location: http://%s:%u" ROOTDESC_PATH "%s" "\r\n"
#else
                 "Location: http://%s:%u" ROOTDESC_PATH  "\r\n"
#endif
                 "Cache-Control: max-age=1209600\r\n"
                 "ST: %.*s\r\n"
                 "USN: %s%s%s\r\n"
                 "\r\n",
#ifdef IFX_DEFINE_TR64
		 host,(unsigned int)port, uuidvalue,
#else
		 host,(unsigned int)port,
#endif
                 st_len,st, 
		 uuid, (iUuidFlag ? "":"::"),(iUuidFlag ? "": st ));
    n = sendto(s, buf, l, 0,
               (struct sockaddr *)&sockname, sizeof(struct sockaddr_in));
    if(n < 0) {
        SYSLOG(LOG_ERR, "sendto(udp): %m");
    }
}

void
SendSSDPNotifies(int s, const char *host, unsigned short port)
{
    struct sockaddr_in sockname;
    int n, i = 0;
    char bufr[512];
    char *DeviceList[200] = { '\0' };
    unsigned int uiCount = 0;
    char sDevUuidVal[50] = {0};
    int iUuidFlag = 0;

    memset(&sockname, 0, sizeof(struct sockaddr_in));
    sockname.sin_family = AF_INET;
    sockname.sin_port = htons(SSDP_PORT);
    sockname.sin_addr.s_addr = inet_addr(SSDP_MCAST_ADDR);

    // Generate the list of actions and services
    IFX_TR64_GenDevList(DeviceList, &uiCount);

    strcpy(sDevUuidVal, uuidvalue);

    while(DeviceList[i]) {
        //change the uuid for the device
        if(strstr(DeviceList[i],"uuid")){
                strcpy(sDevUuidVal, DeviceList[i]);
		iUuidFlag = 1;
        }else{
		iUuidFlag = 0;
	}

        snprintf(bufr, sizeof(bufr),
                 "NOTIFY * HTTP/1.1\r\n"
                 "HOST:%s:%d\r\n"
                 "Cache-Control:max-age=1209600\r\n"
#ifdef IFX_DEFINE_TR64
                 "Location:http://%s:%d" ROOTDESC_PATH "%s" "\r\n"
#else
                 "Location:http://%s:%d" ROOTDESC_PATH"\r\n"
#endif
                 /* "Server:miniupnpd/1.0 UPnP/1.0\r\n" */
                 "Server: " MINIUPNPD_SERVER_STRING "\r\n"
                 "NT:%s\r\n"
                 "USN:%s%s%s\r\n"
                 "NTS:ssdp:alive\r\n"
                 "\r\n",
                 SSDP_MCAST_ADDR, SSDP_PORT,
#ifdef IFX_DEFINE_TR64
                 host, port,uuidvalue,
#else
                 host, port,
#endif
                 DeviceList[i], 
                 sDevUuidVal, (iUuidFlag ? "" : "::"), (iUuidFlag ? "" : DeviceList[i]));
        n = sendto(s, bufr, strlen(bufr), 0,
                   (struct sockaddr *)&sockname, sizeof(struct sockaddr_in));
        if(n < 0) {
            SYSLOG(LOG_ERR, "sendto(udp_notify): %m");
        }
        i++;
    }
    IFX_TR64_FreeDevList(DeviceList);
}

void
ProcessSSDPRequest(int s, const char *host, unsigned short port)
{
    int n;
    char bufr[2048];
    socklen_t len_r;
    struct sockaddr_in sendername;
    int i, l;
    char *st = 0;
    int st_len = 0;
    len_r = sizeof(struct sockaddr_in);
    int index = 0;
    unsigned int uiCount = 0;
    char *DeviceList[90] = {'\0'};
    char sDevUuidVal[50] = {0};
    int iUuidFlag = 0;


    n = recvfrom(s, bufr, sizeof(bufr), 0,
                 (struct sockaddr *)&sendername, &len_r);
    if(n < 0) {
        SYSLOG(LOG_ERR, "recvfrom(udp): %m");
        return;
    }

    if(memcmp(bufr, "NOTIFY", 6) == 0) {
        /* ignore NOTIFY packets. We could log the sender and device type */
        return;
    }
    else if(memcmp(bufr, "M-SEARCH", 8) == 0) {
        i = 0;
        while(i < n) {
            while(bufr[i] != '\r' || bufr[i + 1] != '\n')
                i++;
            i += 2;
            if(strncasecmp(bufr + i, "st:", 3) == 0) {
                st = bufr + i + 3;
                st_len = 0;
                while(*st == ' ' || *st == '\t')
                    st++;
                while(st[st_len] != '\r' && st[st_len] != '\n')
                    st_len++;
                /* SYSLOG(LOG_INFO, "ST: %.*s", st_len, st); */
                /* j = 0; */
                /* while(bufr[i+j]!='\r') j++; */
                /* SYSLOG(LOG_INFO, "%.*s", j, bufr+i); */
            }
        }
        /* SYSLOG(LOG_INFO, "SSDP M-SEARCH packet received from %s:%d",
           inet_ntoa(sendername.sin_addr), ntohs(sendername.sin_port) ); */
	if( (strlen("ssdp:all") <= st_len) && (0 == memcmp(st, "ssdp:all", strlen("ssdp:all")) )){

		IFX_TR64_GenDevList(DeviceList, &uiCount);

		strcpy(sDevUuidVal, uuidvalue);

		while(DeviceList[index]){
			if(strstr(DeviceList[index],"uuid")){
				strcpy(sDevUuidVal, DeviceList[index]);
				iUuidFlag = 1;
			}else{
				iUuidFlag = 0;
			}

			SendSSDPAnnounce2(s, sendername, DeviceList[index], strlen(DeviceList[index]), host, port, sDevUuidVal, iUuidFlag);
			index++;
		}
		
	}else if(st) {
            /* TODO : doesnt answer at once but wait for a random time */
            /* SYSLOG(LOG_INFO, "ST: %.*s", st_len, st); */
            SYSLOG(LOG_INFO, "SSDP M-SEARCH from %s:%d ST: %.*s",
                   inet_ntoa(sendername.sin_addr),
                   ntohs(sendername.sin_port), st_len, st);

	    IFX_TR64_GenDevList(DeviceList, &uiCount);
	
	    strcpy(sDevUuidVal, uuidvalue);


	    index = 0;
	    while(DeviceList[index]) {

		    if(strstr(DeviceList[index],"uuid")){
			    strcpy(sDevUuidVal, DeviceList[index]);
			    iUuidFlag = 1;
		    }else{
			    iUuidFlag = 0;
		    }


		    l = (int)strlen(DeviceList[index]);
		    if(l <= st_len && (strstr(st, DeviceList[index]))) {
			    SendSSDPAnnounce2(s, sendername, st, st_len, host, port, sDevUuidVal, iUuidFlag);
			    break;
		    }
		    index++;
	    }
	    //    l = (int)strlen(uuidvalue);
          //  if(l == st_len && (0 == memcmp(st, uuidvalue, l))) {
            //    SendSSDPAnnounce2(s, sendername, st, st_len, host, port, sDevUuidVal);
           // }
        }
        else {
            SYSLOG(LOG_INFO, "Invalid SSDP M-SEARCH from %s:%d",
                   inet_ntoa(sendername.sin_addr), ntohs(sendername.sin_port));
        }


	IFX_TR64_FreeDevList(DeviceList);
    }
    else {
        SYSLOG(LOG_NOTICE, "Unknown udp packet received from %s:%d",
               inet_ntoa(sendername.sin_addr), ntohs(sendername.sin_port));
    }
}

/* This will broadcast ssdp:byebye notifications to inform
 * the network that UPnP is going down. */
int
SendSSDPGoodbye(int s)
{
    struct sockaddr_in sockname;
    int n;
    int i = 0;
    char bufr[512];
    char *DeviceList[90] = {'\0'};
    unsigned int uiCount = 0;
    char sDevUuidVal[50] = {0};
    int iUuidFlag = 0;

    memset(&sockname, 0, sizeof(struct sockaddr_in));
    sockname.sin_family = AF_INET;
    sockname.sin_port = htons(SSDP_PORT);
    sockname.sin_addr.s_addr = inet_addr(SSDP_MCAST_ADDR);

    // Generate the list of actions and services
    IFX_TR64_GenDevList(DeviceList, &uiCount);

    strcpy(sDevUuidVal, uuidvalue); 

    while(DeviceList[i]) {
	//change the uuid for the device
	if(strstr(DeviceList[i],"uuid")){
		strcpy(sDevUuidVal, DeviceList[i]);
		iUuidFlag = 1;
        }else{
		iUuidFlag = 0;
	}


        snprintf(bufr, sizeof(bufr),
                 "NOTIFY * HTTP/1.1\r\n"
                 "HOST:%s:%d\r\n"
                 "NT:%s\r\n"
                 "USN:%s%s%s\r\n"
                 "NTS:ssdp:byebye\r\n"
                 "\r\n",
                 SSDP_MCAST_ADDR, SSDP_PORT,
                 DeviceList[i], 
                 sDevUuidVal, (iUuidFlag ? "" : "::"), (iUuidFlag ? "" : DeviceList[i]));
        n = sendto(s, bufr, strlen(bufr), 0,
                   (struct sockaddr *)&sockname, sizeof(struct sockaddr_in));
        if(n < 0) {
            SYSLOG(LOG_ERR, "sendto(udp_shutdown): %m");
            return -1;
        }
        i++;
    }

    IFX_TR64_FreeDevList(DeviceList);
    return 0;
}
