///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2000-2003 Intel Corporation 
// All rights reserved. 
//
// Redistribution and use in source and binary forms, with or without 
// modification, are permitted provided that the following conditions are met: 
//
// * Redistributions of source code must retain the above copyright notice, 
// this list of conditions and the following disclaimer. 
// * Redistributions in binary form must reproduce the above copyright notice, 
// this list of conditions and the following disclaimer in the documentation 
// and/or other materials provided with the distribution. 
// * Neither name of Intel Corporation nor the names of its contributors 
// may be used to endorse or promote products derived from this software 
// without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR 
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#ifdef INCLUDE_DEVICE_APIS
#if EXCLUDE_SOAP == 0

#ifdef DEBUG
#define DEBUGONLY(x) x
#else
#define DEBUGONLY(x)
#endif

#define SOAP_BODY "Body"
#define SOAP_URN "http://schemas.xmlsoap.org/soap/envelope/"

#define QUERY_STATE_VAR_URN "urn:schemas-upnp-org:control-1-0"

#include "config.h"
#include "upnpapi.h"
#include "parsetools.h"
#include "statcodes.h"
#include "httpparser.h"
#include "httpreadwrite.h"
#include "unixutil.h"
#include "soaplib.h"

// timeout duration in secs for transmission/reception
#define SOAP_TIMEOUT UPNP_TIMEOUT

#define SREQ_HDR_NOT_FOUND	 -1
#define SREQ_BAD_HDR_FORMAT	 -2

#define SOAP_INVALID_ACTION 	401
#define SOAP_INVALID_ARGS	402
#define SOAP_OUT_OF_SYNC	403
#define SOAP_INVALID_VAR	404
#define SOAP_ACTION_FAILED	501

static const char *Soap_Invalid_Action = "Invalid Action";
//static const char* Soap_Invalid_Args = "Invalid Args";
static const char *Soap_Action_Failed = "Action Failed";
static const char *Soap_Invalid_Var = "Invalid Var";

const char *ContentTypeHeader =
    "Content-Type: text/xml\n\n";

#define		CPE_MAX_FAULT_CODES	14
const faultTbl_t cpeCodes[CPE_MAX_FAULT_CODES] = {
    {9000, "Method not supported", "Server"},
    {9001, "Request denied", "Server"},
    {9002, "Internal Error", "Server"},
    {9003, "Invalid Arguments", "Client"},
    {9004, "Resources exceeded", "Server"},
    {9005, "Invalid Parameter Name", "Client"},
    {9006, "Invalid Parameter Type", "Client"},
    {9007, "Invalid Parameter Value", "Client"},
    {9008, "Attempt to set a non-writable parameter", "Client"},
    {9009, "Notification request rejected", "Server"},
    {9010, "Download failure", "Server"},
    {9011, "Upload failure", "Server"},
    {9012, "File transfer server authentication failure", "Server"},
    {9013, "Unsupported protocol for file transfer", "Server"},
};

#define 	CPE_RPC_CALLBACK_MAX	11
callbackTbl_t cb[CPE_RPC_CALLBACK_MAX] = {
    {"GetRPCMethods", NULL},
    {"Fault", NULL},
    {"InformResponse", NULL},
    {"GetParameterValues", NULL},
    {"SetParameterValues", NULL},
    {"GetParameterAttributes", NULL},
    {"SetParameterAttributes", NULL},
    {"GetParameterNames", NULL},
    {"Addobject", NULL},
    {"DeleteObject", NULL},
    {"Download", NULL},
};

static char *strEnv = 
#ifdef _TR69
		"<?xml version=\"1.0\"?>"
#endif
		"<SOAP-ENV:Envelope xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"SOAP-ENV:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" "
//		"xmlns:SOAP-ENC=\"http://schemas.xmlsoap.org/soap/encoding/\" "
		"xmlns:xsd=\"http://www.w3.org/2001/XMLSchema-instance\" "
		"xmlns:xsi=\"http://www.w3.org/2001/XMLSchema\" "
		"xmlns:cwmp=\"urn:dslforum-org:cwmp-1-0\">";
static char *strHeadStart = "<SOAP-ENV:Header>";
static char *strHeadEnd = "</SOAP-ENV:Header>";
static char *strIDStart = "<cwmp:ID SOAP-ENV:mustUnderstand=\"1\">";
static char *strIDEnd = "</cwmp:ID>";
static char *strBodyStart = "<SOAP-ENV:Body>";
static char *strBodyEnd = "</SOAP-ENV:Body></SOAP-ENV:Envelope>";
SOCKINFO *sockinfo = NULL;


extern int get_port(int socket);

/****************************************************************************
*	Function :	get_request_type
*
*	Parameters :
*			IN http_message_t* request :	HTTP request
*			OUT memptr* action_name :	SOAP action name
*
*	Description :	This function retrives the name of the SOAP action
*
*	Return : int
*		0 if successful else returns appropriate error.
*	Note :
****************************************************************************/
static XINLINE int
get_request_type( IN http_message_t * request,
                  OUT memptr * action_name )
{
    memptr value;
    memptr ns_value,
      dummy_quote;
    http_header_t *hdr;
    char save_char;
    char *s;
    membuffer soap_action_name;

    // find soapaction header
    //
    if( request->method == SOAPMETHOD_POST ) {
        if( httpmsg_find_hdr( request, HDR_SOAPACTION, &value )
            == NULL ) {
            return SREQ_HDR_NOT_FOUND;
        }
    } else                      // M-POST
    {
        // get NS value from MAN header
        hdr = httpmsg_find_hdr( request, HDR_MAN, &value );
        if( hdr == NULL ) {
            return SREQ_HDR_NOT_FOUND;
        }

        if( matchstr( value.buf, value.length, "%q%i ; ns = %s",
                      &dummy_quote, &ns_value ) != 0 ) {
            return SREQ_BAD_HDR_FORMAT;
        }
        // create soapaction name header
        membuffer_init( &soap_action_name );
        if( ( membuffer_assign( &soap_action_name,
                                ns_value.buf, ns_value.length )
              == UPNP_E_OUTOF_MEMORY ) ||
            ( membuffer_append_str( &soap_action_name,
                                    "-SOAPACTION" ) ==
              UPNP_E_OUTOF_MEMORY )
             ) {
            membuffer_destroy( &soap_action_name );
            return UPNP_E_OUTOF_MEMORY;
        }

        hdr = httpmsg_find_hdr_str( request, soap_action_name.buf );
        membuffer_destroy( &soap_action_name );
        if( hdr == NULL ) {
            return SREQ_HDR_NOT_FOUND;
        }

        value.buf = hdr->value.buf;
        value.length = hdr->value.length;
    }

    // determine type
    //
    save_char = value.buf[value.length];
    value.buf[value.length] = '\0';

    s = strchr( value.buf, '#' );
    if( s == NULL ) {
        value.buf[value.length] = save_char;
        return SREQ_BAD_HDR_FORMAT;
    }

    s++;                        // move to value

    if( matchstr( s, value.length - ( s - value.buf ), "%s",
                  action_name ) != PARSE_OK ) {
        value.buf[value.length] = save_char;
        return SREQ_BAD_HDR_FORMAT;
    }
    // action name or variable ?
    if( memptr_cmp( action_name, "QueryStateVariable" ) == 0 ) {
        // query variable
        action_name->buf = NULL;
        action_name->length = 0;
    }

    value.buf[value.length] = save_char;    // restore
    return 0;
}

/****************************************************************************
*	Function :	send_error_response
*
*	Parameters :
*			IN SOCKINFO *info :	socket info
*			IN int error_code :	error code
*			IN const char* err_msg :	error message
*			IN http_message_t* hmsg :	HTTP request
*
*	Description :	This function sends SOAP error response
*
*	Return : void
*
*	Note :
****************************************************************************/
static void
send_error_response( IN SOCKINFO * info,
                     IN int error_code,
                     IN const char *err_msg,
                     IN http_message_t * hmsg )
{
    int content_length;
    int timeout_secs = SOAP_TIMEOUT;
    int major,
      minor;
    const char *start_body =
        "<s:Envelope\n"
        "xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\"\n"
        "s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">\n"
        "<s:Body>\n"
        "<s:Fault>\n"
        "<faultcode>s:Client</faultcode>\n"
        "<faultstring>UPnPError</faultstring>\n"
        "<detail>\n"
        "<UPnPError xmlns=\"urn:schemas-upnp-org:control-1-0\">\n"
        "<errorCode>";

    const char *mid_body = "</errorCode>\n" "<errorDescription>";

    const char *end_body =
        "</errorDescription>\n"
        "</UPnPError>\n"
        "</detail>\n" "</s:Fault>\n" "</s:Body>\n" "</s:Envelope>\n";

    char err_code_str[30];

    membuffer headers;

    sprintf( err_code_str, "%d", error_code );

    // calc body len
    content_length = strlen( start_body ) + strlen( err_code_str ) +
        strlen( mid_body ) + strlen( err_msg ) + strlen( end_body );

    http_CalcResponseVersion( hmsg->major_version, hmsg->minor_version,
                              &major, &minor );

    // make headers
    membuffer_init( &headers );
    if( http_MakeMessage( &headers, major, minor,
                          "RNsDsSc" "sssss",
                          500,
                          content_length,
                          ContentTypeHeader,
                          "EXT:\r\n",
                          start_body, err_code_str, mid_body, err_msg,
                          end_body ) != 0 ) {
        membuffer_destroy( &headers );
        return;                 // out of mem
    }
    // send err msg
    http_SendMessage( info, &timeout_secs, "b",
                      headers.buf, headers.length );

    membuffer_destroy( &headers );
}

/****************************************************************************
*	Function :	send_var_query_response
*
*	Parameters :
*			IN SOCKINFO *info :	socket info
*			IN const char* var_value :	value of the state variable
*			IN http_message_t* hmsg :	HTTP request
*
*	Description :	This function sends response of get var status
*
*	Return : void
*
*	Note :
****************************************************************************/
static XINLINE void
send_var_query_response( IN SOCKINFO * info,
                         IN const char *var_value,
                         IN http_message_t * hmsg )
{
    int content_length;
    int timeout_secs = SOAP_TIMEOUT;
    int major,
      minor;
    const char *start_body =
        "<s:Envelope\n"
        "xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\"\n"
        "s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">\n"
        "<s:Body>\n"
        "<u:QueryStateVariableResponse "
        "xmlns:u=\"urn:schemas-upnp-org:control-1-0\">\n" "<return>";

    const char *end_body =
        "</return>\n"
        "</u:QueryStateVariableResponse>\n"
        "</s:Body>\n" "</s:Envelope>\n";

    membuffer response;

    http_CalcResponseVersion( hmsg->major_version, hmsg->minor_version,
                              &major, &minor );

    content_length = strlen( start_body ) + strlen( var_value ) +
        strlen( end_body );

    // make headers
    membuffer_init( &response );
    if( http_MakeMessage( &response, major, minor,
                          "RNsDsSc" "sss",
                          HTTP_OK,
                          content_length,
                          ContentTypeHeader,
                          "EXT:\r\n",
                          start_body, var_value, end_body ) != 0 ) {
        membuffer_destroy( &response );
        return;                 // out of mem
    }
    // send msg
    http_SendMessage( info, &timeout_secs, "b",
                      response.buf, response.length );

    membuffer_destroy( &response );
}

/****************************************************************************
*	Function :	get_action_node
*
*	Parameters :
*		IN IXML_Document *TempDoc :	The root DOM node.
*		IN char *NodeName :	IXML_Node name to be searched.
*		OUT IXML_Document ** RespNode :	Response/Output node.
*
*	Description :	This function separates the action node from 
*	the root DOM node.
*
*	Return :	static XINLINE int
*		0 if successful, or -1 if fails.
*
*	Note :
****************************************************************************/
static XINLINE int
get_action_node( IN IXML_Document * TempDoc,
                 IN char *NodeName,
                 OUT IXML_Document ** RespNode )
{
    IXML_Node *EnvpNode = NULL;
    IXML_Node *BodyNode = NULL;
    IXML_Node *ActNode = NULL;
    DOMString ActNodeName = NULL;
    const DOMString nodeName;
    int ret_code = -1;          // error, by default
    IXML_NodeList *nl = NULL;

    DBGONLY( UpnpPrintf( UPNP_INFO, SOAP, __FILE__, __LINE__,
                         "get_action_node(): node name =%s\n ", NodeName );
         )

        * RespNode = NULL;

    // Got the Envelope node here
    EnvpNode = ixmlNode_getFirstChild( ( IXML_Node * ) TempDoc );
    if( EnvpNode == NULL ) {
        goto error_handler;
    }

    nl = ixmlElement_getElementsByTagNameNS( ( IXML_Element * ) EnvpNode,
                                             "*", "Body" );

    if( nl == NULL ) {
        goto error_handler;
    }

    BodyNode = ixmlNodeList_item( nl, 0 );

    if( BodyNode == NULL ) {
        goto error_handler;
    }
    // Got action node here
    ActNode = ixmlNode_getFirstChild( BodyNode );
    if( ActNode == NULL ) {
        goto error_handler;
    }
    //Test whether this is the action node
    nodeName = ixmlNode_getNodeName( ActNode );
    if( nodeName == NULL ) {
        goto error_handler;
    }

    if( strstr( nodeName, NodeName ) == NULL ) {
        goto error_handler;
    } else {
        ActNodeName = ixmlPrintDocument( ActNode );
        if( ActNodeName == NULL ) {
            goto error_handler;
        }

        ret_code = ixmlParseBufferEx( ActNodeName, RespNode );
        if( ret_code != IXML_SUCCESS ) {
            ixmlFreeDOMString( ActNodeName );
            ret_code = -1;
            goto error_handler;
        }
    }

    ret_code = 0;               // success

  error_handler:

    ixmlFreeDOMString( ActNodeName );

    if( nl )
        ixmlNodeList_free( nl );
    return ret_code;
}

/****************************************************************************
*	Function :	check_soap_body
*
*	Parameters :
*		IN IXML_Document *doc :	soap body xml document
*		    IN const char *urn : 
*		    IN const char *actionName : Name of the requested action 	
*
*	Description :	This function checks the soap body xml came in the
*		SOAP request.
*
*	Return : int
*		UPNP_E_SUCCESS if successful else returns appropriate error
*
*	Note :
****************************************************************************/
static int
check_soap_body( IN IXML_Document * doc,
                 IN const char *urn,
                 IN const char *actionName )
{
    IXML_NodeList *nl = NULL;
    IXML_Node *bodyNode = NULL;
    IXML_Node *actionNode = NULL;
    const DOMString ns = NULL;
    const DOMString name = NULL;

    int ret_code = UPNP_E_INVALID_ACTION;
    
    nl = ixmlDocument_getElementsByTagNameNS( doc, SOAP_URN, SOAP_BODY );

    if( nl ) {
        bodyNode = ixmlNodeList_item( nl, 0 );
        if( bodyNode ) {
            actionNode = ixmlNode_getFirstChild( bodyNode );
            if( actionNode ) {
                ns = ixmlNode_getNamespaceURI( actionNode );
                name = ixmlNode_getLocalName( actionNode );
printf("check_soap_body(), actionname=%s, name=%s\n", actionName, name); //Nirint
printf("check_soap_body(), urn=%s\n, ns=%s\n", urn, ns); //Nirint
                if( ( !strcmp( actionName, name ) )
                    && ( !strcmp( urn, ns ) ) ) {
                    ret_code = UPNP_E_SUCCESS;
                }
            }
        }
        ixmlNodeList_free( nl );
    }
    return ret_code;

}

/****************************************************************************
*	Function :	check_soap_action_header
*
*	Parameters :
*		IN http_message_t *request : HTTP request
*		IN const char *urn :
*		OUT char **actionName :	 name of the SOAP action
*
*	Description :	This function checks the HTTP header of the SOAP request
*		coming from the control point
*
*	Return :	static int
*		UPNP_E_SUCCESS if successful else returns appropriate error
*
*	Note :
****************************************************************************/
static int
check_soap_action_header( IN http_message_t * request,
                          IN const char *urn,
                          OUT char **actionName )
{
    memptr header_name;
    http_header_t *soap_action_header = NULL;
    char *ns_compare = NULL;
    int tempSize = 0;
    int ret_code = UPNP_E_SUCCESS;
    char *temp_header_value = NULL;
    char *temp = NULL;
    char *temp2 = NULL;

    //check soap action header

    soap_action_header = httpmsg_find_hdr( request, HDR_SOAPACTION,
                                           &header_name );
    if( !soap_action_header ) {
        ret_code = UPNP_E_INVALID_ACTION;
        return ret_code;
    }

    if( soap_action_header->value.length <= 0 ) {
        ret_code = UPNP_E_INVALID_ACTION;
        return ret_code;
    }

    temp_header_value =
        ( char * )malloc( soap_action_header->value.length + 1 );

    if( !temp_header_value ) {
        ret_code = UPNP_E_OUTOF_MEMORY;
        free( temp_header_value );
        return ret_code;
    }

    strncpy( temp_header_value, soap_action_header->value.buf,
             soap_action_header->value.length );
    temp_header_value[soap_action_header->value.length] = 0;

    temp = strchr( temp_header_value, '#' );
    if( !temp ) {
        free( temp_header_value );
        ret_code = UPNP_E_INVALID_ACTION;
        return ret_code;
    }

    ( *temp ) = 0;              //temp make string

    //check to see if it is Query State Variable or
    //Service Action

    tempSize = strlen( urn ) + 2;

    ns_compare = ( char * )malloc( tempSize );

    if( !ns_compare ) {
        ret_code = UPNP_E_OUTOF_MEMORY;
        free( temp_header_value );
        return ret_code;
    }

    //snprintf( ns_compare, tempSize, "\"%s", urn );
//Jeffrey_0924+
    if (temp_header_value[0] == '\"')
        snprintf(ns_compare, tempSize, "\"%s", urn);
    else
        snprintf(ns_compare, tempSize, "%s", urn);


    //printf("check_soap_action_header 4 tmp_header_value=%s,\nns_compare=%s\n", temp_header_value, ns_compare); //Nirint
    if( strcmp( temp_header_value, ns_compare ) ) {
        ret_code = UPNP_E_INVALID_ACTION;
    } else {
        ret_code = UPNP_E_SUCCESS;
        temp++;
        temp2 = strchr( temp, '\"' );

        if( temp2 )             //remove ending " if present
        {
            ( *temp2 ) = 0;
        }

        if( *temp )
            ( *actionName ) = strdup( temp );
        if( !*actionName ) {
            ret_code = UPNP_E_OUTOF_MEMORY;
        }
    }

    free( temp_header_value );
    free( ns_compare );
    return ret_code;
}

/****************************************************************************
*	Function :	get_device_info
*
*	Parameters :
*		IN http_message_t* request :	HTTP request
*		IN int isQuery :	flag for a querry
*		IN IXML_Document *actionDoc :	action request document
*		OUT char device_udn[LINE_SIZE] :	Device UDN string
*		OUT char service_id[LINE_SIZE] :	Service ID string
*		OUT Upnp_FunPtr *callback :	callback function of the device 
*									application
*		OUT void** cookie :	cookie stored by device application 
*							
*	Description :	This function retrives all the information needed to 
*		process the incoming SOAP request. It finds the device and service info
*		and also the callback function to hand-over the request to the device
*		application.
*
*	Return : int
*		UPNP_E_SUCCESS if successful else returns appropriate error
*
*	Note :
****************************************************************************/
static int
get_device_info( IN http_message_t * request,
                 IN int isQuery,
                 IN IXML_Document * actionDoc,
                 OUT char device_udn[LINE_SIZE],
                 OUT char service_id[LINE_SIZE],
                 OUT Upnp_FunPtr * callback,
                 OUT void **cookie )
{
    struct Handle_Info *device_info;
    int device_hnd;
    service_info *serv_info;
    char save_char;
    int ret_code = -1;          // error by default
    char *control_url;
    char *actionName = NULL;

    // null-terminate pathquery of url
    control_url = request->uri.pathquery.buff;
    save_char = control_url[request->uri.pathquery.size];
    control_url[request->uri.pathquery.size] = '\0';

    HandleLock(  );

    //printf("get_device_info() before getdevicehandle\n");
    if( GetDeviceHandleInfo( &device_hnd, &device_info ) != HND_DEVICE ) {
        goto error_handler;
    }

    //printf("get_device_info() before findservicectrl %s\n", control_url);
    if( ( serv_info =
          FindServiceControlURLPath( &device_info->ServiceTable,
                                     control_url ) ) == NULL ) {
        goto error_handler;
    }
    //printf("get_device_info() before checksoapaction\n");

    if( isQuery ) {
        ret_code = check_soap_action_header( request, QUERY_STATE_VAR_URN,
                                             &actionName );
        if( ( ret_code != UPNP_E_SUCCESS )
            && ( ret_code != UPNP_E_OUTOF_MEMORY ) ) {
            ret_code = UPNP_E_INVALID_ACTION;
            goto error_handler;
        }
        //check soap body
        ret_code =
            check_soap_body( actionDoc, QUERY_STATE_VAR_URN, actionName );
        free( actionName );
        if( ret_code != UPNP_E_SUCCESS ) {
	    printf("error!! check_soap_body\n"); // Nirint
//            goto error_handler;
        } 
    } else {
        ret_code = check_soap_action_header( request,
                                             serv_info->serviceType,
                                             &actionName );
        if( ( ret_code != UPNP_E_SUCCESS )
            && ( ret_code != UPNP_E_OUTOF_MEMORY ) ) {
            ret_code = UPNP_E_INVALID_SERVICE;
            goto error_handler;
        }
        //check soap body
        ret_code =
            check_soap_body( actionDoc, serv_info->serviceType,
                             actionName );
        free( actionName );
        if( ret_code != UPNP_E_SUCCESS ) {
            ret_code = UPNP_E_INVALID_SERVICE;
            goto error_handler;
        }
    }

    namecopy( service_id, serv_info->serviceId );
    namecopy( device_udn, serv_info->UDN );
    *callback = device_info->Callback;
    *cookie = device_info->Cookie;

    ret_code = 0;

  error_handler:
    control_url[request->uri.pathquery.size] = save_char;   // restore
    HandleUnlock(  );
    return ret_code;
}

/****************************************************************************
*	Function :	send_action_response
*
*	Parameters :
*		IN SOCKINFO *info :	socket info
*		IN IXML_Document *action_resp : The response document	
*		IN http_message_t* request :	action request document
*
*	Description :	This function sends the SOAP response 
*
*	Return : void
*
*	Note :
****************************************************************************/
static XINLINE void
send_action_response( IN SOCKINFO * info,
                      IN IXML_Document * action_resp,
                      IN http_message_t * request )
{
    char *xml_response = NULL;
    membuffer headers;
    int major,
      minor;
    int err_code;
    int content_length;
    int ret_code;
    int timeout_secs = SOAP_TIMEOUT;
    static char *start_body =
        "<s:Envelope xmlns:s=\"http://schemas.xmlsoap."
        "org/soap/envelope/\" s:encodingStyle=\"http://schemas.xmlsoap."
        "org/soap/encoding/\"><s:Body>\n";
    static char *end_body = "</s:Body> </s:Envelope>";

    // init
    http_CalcResponseVersion( request->major_version,
                              request->minor_version, &major, &minor );
    membuffer_init( &headers );
    err_code = UPNP_E_OUTOF_MEMORY; // one error only

    // get xml
    xml_response = ixmlPrintDocument( ( IXML_Node * ) action_resp );
    if( xml_response == NULL ) {
        goto error_handler;
    }

    content_length = strlen( start_body ) + strlen( xml_response ) +
        strlen( end_body );

    // make headers
    if( http_MakeMessage( &headers, major, minor, "RNsDsSc", HTTP_OK,   // status code
                          content_length, ContentTypeHeader, "EXT:\r\n" // EXT header
         ) != 0 ) {
        goto error_handler;
    }
    // send whole msg
    ret_code = http_SendMessage( info, &timeout_secs, "bbbb",
                                 headers.buf, headers.length,
                                 start_body, strlen( start_body ),
                                 xml_response, strlen( xml_response ),
                                 end_body, strlen( end_body ) );

    DBGONLY( if( ret_code != 0 ) {
             UpnpPrintf( UPNP_INFO, SOAP, __FILE__, __LINE__,
                         "Failed to send response: err code = %d\n",
                         ret_code );}
     )

        err_code = 0;

  error_handler:
    ixmlFreeDOMString( xml_response );
    membuffer_destroy( &headers );
    if( err_code != 0 ) {
        // only one type of error to worry about - out of mem
        send_error_response( info, SOAP_ACTION_FAILED, "Out of memory",
                             request );
    }
}

/****************************************************************************
*	Function :	get_var_name
*
*	Parameters :
*		IN IXML_Document *TempDoc :	Document containing variable request
*		OUT char* VarName :	Name of the state varible
*
*	Description :	This function finds the name of the state variable 
*				asked in the SOAP request.
*
*	Return :	int
*		returns 0 if successful else returns -1.
*	Note :
****************************************************************************/
static XINLINE int
get_var_name( IN IXML_Document * TempDoc,
              OUT char *VarName )
{
    IXML_Node *EnvpNode = NULL;
    IXML_Node *BodyNode = NULL;
    IXML_Node *StNode = NULL;
    IXML_Node *VarNameNode = NULL;
    IXML_Node *VarNode = NULL;
    const DOMString StNodeName = NULL;
    DOMString Temp = NULL;
    int ret_val = -1;

    // Got the Envelop node here
    EnvpNode = ixmlNode_getFirstChild( ( IXML_Node * ) TempDoc );
    if( EnvpNode == NULL ) {
        goto error_handler;
    }
    //printf("soap get var name0 nodename%s value%s\n",ixmlNode_getNodeName(EnvpNode), ixmlNode_getNodeValue(EnvpNode)); // Nirint
    // Got Body here
    BodyNode = ixmlNode_getFirstChild( EnvpNode );
    if( BodyNode == NULL ) {
        goto error_handler;
    }
    //printf("soap get var name1 nodename%s value%s\n",ixmlNode_getNodeName(BodyNode), ixmlNode_getNodeValue(BodyNode)); // Nirint
    // Got action node here
    StNode = ixmlNode_getFirstChild( BodyNode );
    if( StNode == NULL ) {
        goto error_handler;
    }
    //printf("soap get var name2 Namespac%s Local%s\n",ixmlNode_getNamespaceURI(StNode),ixmlNode_getLocalName(StNode)); // Nirint
    //Test whether this is the action node
    StNodeName = ixmlNode_getNodeName( StNode );
    //printf("soap get var name3 nodename%s value%s\n",StNodeName, ixmlNode_getNodeValue(StNodeName)); // Nirint
    if( StNodeName == NULL || strstr( StNodeName,
                                      "QueryStateVariable" ) == NULL ) {
        goto error_handler;
    }

    VarNameNode = ixmlNode_getFirstChild( StNode );
    if( VarNameNode == NULL ) {
        goto error_handler;
    }
    printf("soap get var name 4 %s\n", ixmlNode_getNodeName(VarNameNode)); // Nirint

    VarNode = ixmlNode_getFirstChild( VarNameNode );
    printf("soap get var name 5 %s\n", ixmlNode_getNodeName(VarNode)); // Nirint
    Temp = ixmlNode_getNodeValue( VarNode );
    printf("soap get var name 6 %s\n", Temp); // Nirint
    linecopy( VarName, Temp );

    DBGONLY( UpnpPrintf( UPNP_INFO, SOAP, __FILE__, __LINE__,
                         "Received query for variable  name %s\n",
                         VarName );
         )

        ret_val = 0;            // success

  error_handler:
    return ret_val;
}

/****************************************************************************
*	Function :	handle_query_variable
*
*	Parameters :
*		IN SOCKINFO *info :	Socket info
*		IN http_message_t* request : HTTP request	
*		IN IXML_Document *xml_doc :	Document containing the variable request 
*									SOAP message
*
*	Description :	This action handles the SOAP requests to querry the 
*				state variables. This functionality has been deprecated in 
*				the UPnP V1.0 architecture
*
*	Return :	void
*
*	Note :
****************************************************************************/
static XINLINE void
handle_query_variable( IN SOCKINFO * info,
                       IN http_message_t * request,
                       IN IXML_Document * xml_doc )
{
    Upnp_FunPtr soap_event_callback;
    void *cookie;
    char var_name[LINE_SIZE];
    struct Upnp_State_Var_Request variable;
    const char *err_str;
    int err_code;

    // get var name
    if( get_var_name( xml_doc, var_name ) != 0 ) {
        send_error_response( info, SOAP_INVALID_VAR,
                             Soap_Invalid_Var, request );
        return;
    }
    // get info for event
    if( get_device_info( request, 1, xml_doc, variable.DevUDN,
                         variable.ServiceID,
                         &soap_event_callback, &cookie ) != 0 ) {
        send_error_response( info, SOAP_INVALID_VAR,
                             Soap_Invalid_Var, request );
        return;
    }
    printf("handle_query_variable(), after getting device info\n"); // Nirint
    linecopy( variable.ErrStr, "" );
    variable.ErrCode = UPNP_E_SUCCESS;
    namecopy( variable.StateVarName, var_name );
    variable.CurrentVal = NULL;
    variable.CtrlPtIPAddr = info->foreign_ip_addr;

    // send event
    soap_event_callback( UPNP_CONTROL_GET_VAR_REQUEST, &variable, cookie );

    DBGONLY( UpnpPrintf( UPNP_INFO, SOAP, __FILE__, __LINE__,
                         "Return from callback for var request\n" ) );

    // validate, and handle result
    if( variable.CurrentVal == NULL ) {
        err_code = SOAP_ACTION_FAILED;
        err_str = Soap_Action_Failed;
        send_error_response( info, SOAP_INVALID_VAR,
                             Soap_Invalid_Var, request );
        return;
    }
    if( variable.ErrCode != UPNP_E_SUCCESS ) {
        if( strlen( variable.ErrStr ) > 0 ) {
            err_code = SOAP_INVALID_VAR;
            err_str = Soap_Invalid_Var;
        } else {
            err_code = variable.ErrCode;
            err_str = variable.ErrStr;
        }
        send_error_response( info, err_code, err_str, request );
        return;
    }
    // send response
    send_var_query_response( info, variable.CurrentVal, request );
    ixmlFreeDOMString( variable.CurrentVal );

}

/****************************************************************************
*	Function :	handle_invoke_action
*
*	Parameters :
*		IN SOCKINFO *info :	Socket info
*		IN http_message_t* request : HTTP Request	
*		IN memptr action_name :	 Name of the SOAP Action
*		IN IXML_Document *xml_doc :	document containing the SOAP action 
*									request
*
*	Description :	This functions handle the SOAP action request. It checks 
*		the integrity of the SOAP action request and gives the call back to 
*		the device application.
*
*	Return : void
*
*	Note :
****************************************************************************/
static void
handle_invoke_action( IN SOCKINFO * info,
                      IN http_message_t * request,
                      IN memptr action_name,
                      IN IXML_Document * xml_doc )
{
    char save_char;
    IXML_Document *resp_node = NULL;
    struct Upnp_Action_Request action;
    Upnp_FunPtr soap_event_callback;
    void *cookie = NULL;
    int err_code;
    const char *err_str;

    action.ActionResult = NULL;

    // null-terminate
    save_char = action_name.buf[action_name.length];
    action_name.buf[action_name.length] = '\0';

    // set default error
    err_code = SOAP_INVALID_ACTION;
    err_str = Soap_Invalid_Action;

    // get action node
    if( get_action_node( xml_doc, action_name.buf, &resp_node ) == -1 ) {
        goto error_handler;
    }
    // get device info for action event
    err_code = get_device_info( request, 0, xml_doc, action.DevUDN,
                                action.ServiceID, &soap_event_callback,
                                &cookie );

    if( err_code != UPNP_E_SUCCESS ) {
        goto error_handler;
    }

    namecopy( action.ActionName, action_name.buf );
    linecopy( action.ErrStr, "" );
    action.ActionRequest = resp_node;
    action.ActionResult = NULL;
    action.ErrCode = UPNP_E_SUCCESS;
    action.CtrlPtIPAddr = info->foreign_ip_addr;

    DBGONLY( UpnpPrintf( UPNP_INFO, SOAP, __FILE__, __LINE__,
                         "Calling Callback\n" ) );

    soap_event_callback( UPNP_CONTROL_ACTION_REQUEST, &action, cookie );

    if( action.ErrCode != UPNP_E_SUCCESS ) {
        if( strlen( action.ErrStr ) <= 0 ) {
            err_code = SOAP_ACTION_FAILED;
            err_str = Soap_Action_Failed;
        } else {
            err_code = action.ErrCode;
            err_str = action.ErrStr;
        }
        goto error_handler;
    }
    // validate, and handle action error
    if( action.ActionResult == NULL ) {
        err_code = SOAP_ACTION_FAILED;
        err_str = Soap_Action_Failed;
        goto error_handler;
    }
    // send response
    send_action_response( info, action.ActionResult, request );

    err_code = 0;

    // error handling and cleanup
  error_handler:
    ixmlDocument_free( action.ActionResult );
    ixmlDocument_free( resp_node );
    action_name.buf[action_name.length] = save_char;    // restore
    if( err_code != 0 ) {
        send_error_response( info, err_code, err_str, request );
    }
}

int ifx_cwmp_send_rpc_resp(SOCKINFO *info, char *rpcResp, char *id, int isInform)
{
    int content_length;
    membuffer headers;
    int timeout_secs = SOAP_TIMEOUT;

    const char *url_str = "http://192.168.1.2:9995/";
    void *Handle;
    char *contentType = NULL;
    int contentLength=0;
    int httpStatus;
    int lowRange=1;
    int highRange=100;
    int status;
    int timeout=18;
    //membuffer request;
    //uri_type url;

    //printf("BUF: %s\n", rpcResp);
    membuffer_init( &headers );
    content_length = strlen(strEnv) + strlen(strBodyStart) + strlen(rpcResp) + strlen(strBodyEnd);
    if((id != NULL) && (atoi(id) > 0)) {
	printf("id not null\n");
	content_length += strlen(strHeadStart) + strlen(strHeadEnd) + strlen(strIDStart) + 
				strlen(strIDEnd) + strlen(id);
        if(http_MakeMessage(&headers, 1, 1, "sNs" "sssssssss", "POST / HTTP/1.1\n", content_length, 
		ContentTypeHeader, strEnv, strHeadStart, strIDStart, id, 
		strIDEnd, strHeadEnd, strBodyStart, rpcResp, strBodyEnd) != 0 ) {
            membuffer_destroy( &headers );
	    printf("ifx_cwmp_send_rpc_response(): Kernel Panic!! Out of memory. Line=%d",__LINE__);
            exit(-1);
        }
    }
    else {
	printf("null id\n");
	if(http_MakeMessage( &headers, 1, 1, "sNs" "ssss", "POST / HTTP/1.1\n", content_length, 
				ContentTypeHeader, strEnv, 
				strBodyStart, rpcResp, strBodyEnd) != 0 ) {
	    membuffer_destroy(&headers);
	    printf("ifx_cwmp_send_rpc_response(): Kernel Panic!! Out of memory. Line=%d",__LINE__);
	    exit(-1);
	}
    }
    write(2, "http_SendMessage called\n", 25);
    http_SendMessage( info, &timeout_secs, "b", headers.buf, headers.length );

    //printf("BUF: %s\n", headers.buf);

    membuffer_destroy( &headers );
    return 0;
}

int ifx_cwmp_send_ack()
{
    char *str = "POST / HTTP/1.1\nHost: 192.168.1.2:9995\n\n";
    char *resp, strPort[9];
    int timeout_secs = SOAP_TIMEOUT;
    int port;
    SOCKINFO *info = ifx_cwmp_get_sockinfo();

    port = get_port(info->socket);
    sprintf(strPort,"%i\n", port);
    
    resp = (char *)malloc(strlen(str) + strlen(strPort) + 1);
    strcpy(resp, str);
    //strcat(resp, strPort);
    if(http_SendMessage(info, &timeout_secs, "b", resp, strlen(resp)) != 0) {
	printf("Unable to send 200 OK ack\n");
	return -1;
    }
    free(resp);
    return 0;
}

int ifx_cwmp_send_fault_response(SOCKINFO *info, int err_code, 
					struct paramFault *pf, int num)
{
    int content_length;
    int timeout_secs = SOAP_TIMEOUT;
    char err_code_str[30];
    int i, j;
    membuffer headers;
    char *start_body =
        "<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\""
        "s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
        " <s:Body> <s:Fault> <faultcode>";
    char *start_body_cont = "</faultcode>"
        " <faultstring>CWMP fault</faultstring>"
        " <detail> <cwmp:Fault> <FaultCode>";
    char *mid_body = "</FaultCode> <FaultString>";
    char *mid_body_cont = "<FaultString/>";
    char *end_body = "</cwmp:Fault> </detail> </s:Fault> </s:Body> </s:Envelope>";
    char *param_start = " <SetParameterValuesFault> <ParameterName>";
    char *param_mid = "</ParameterName> <FaultCode>";
    char *param_mid_cont = "</FaultCode> <FaultString>";
    char *param_end = "</FaultString> </SetParameterValuesFault>";
    char *param = NULL;

    for(i = 0; i < CPE_MAX_FAULT_CODES; i++) {
	if(cpeCodes[i].faultCode == err_code) {
	    strcat(start_body, cpeCodes[i].type);
	    strcat(mid_body, cpeCodes[i].faultDesc);
	    break;
	}
    }
    sprintf( err_code_str, "%d", err_code );
    strcat(start_body_cont, err_code_str);

    if(num & (num <= 3)) { //num > 3 will consume too much memory
	param = (char *)malloc(700); //Magic number should be enough to store 3 fault msgs
	if(param == NULL) {
	    printf("ifx_cwmp_send_fault_response(): Kernel Panic!! malloc failed");
	    exit(-1);
	}
	for(j = 0; j < num; j++) {
	    strcat(param_start, (pf+i)->paramName);
	    sprintf(err_code_str, "%d", (pf+i)->faultCode);
	    strcat(param_mid, err_code_str);
	    for(i = 0; i < CPE_MAX_FAULT_CODES; i++) {
		if(cpeCodes[i].faultCode == (pf+i)->faultCode) {
		    strcat(param_mid_cont, cpeCodes[i].faultDesc);
		    break;
		}
	    }
	    strcat(param, param_start);
	    strcat(param, param_mid);
	    strcat(param, param_mid_cont);
	    strcat(param, param_end);
	}
    }

    // calc body len
    content_length = strlen( start_body ) +  strlen( start_body_cont ) + strlen( mid_body ) + 
				strlen( mid_body_cont ) + strlen( end_body ) + strlen( param );

    // make headers
    membuffer_init( &headers );
    if( http_MakeMessage( &headers, 1, 1, "RNsDsSc" "sssss", 500, content_length, 
				ContentTypeHeader, "EXT:\r\n", start_body, start_body_cont, 
				mid_body, mid_body_cont, end_body, param ) != 0 ) {
        membuffer_destroy( &headers );
	printf("ifx_cwmp_send_fault_response(): Kernel Panic!! Out of memory");
        exit(-1);                 // out of mem
    }
    // send fault msg
    http_SendMessage( info, &timeout_secs, "b",
                      headers.buf, headers.length );

    membuffer_destroy( &headers );

    return SOAP_SUCCESS;
}

/*
 *
 *
 */

int ifx_cwmp_register_soap_callback(char *rpcName, func funcptr)
{
    int i;
    for(i = 0; i < CPE_RPC_CALLBACK_MAX; i++) {
	if(strcmp(cb[i].rpcName, rpcName) == 0) {
	    cb[i].funcptr = funcptr;
	    break;
	}
    }
    if(i == CPE_RPC_CALLBACK_MAX)
	return SOAP_FAILURE;
    return SOAP_SUCCESS;
}

/*
 *
 *
 */

int ifx_cwmp_process_soap_body(SOCKINFO *info, IXML_Node *node, char *id)
{
    IXML_Node *childnode = NULL;
    char *nodeName = NULL, *rpcResp = NULL;
    int i, ret;

    printf("Body1: %s\n", ixmlNode_getNodeName(node));
    childnode = ixmlNode_getFirstChild(node);
    nodeName = (char *)ixmlNode_getNodeName(childnode);
    printf("Body2: %s\n", nodeName);
    if(nodeName != NULL)
	for(i = 0; i < CPE_RPC_CALLBACK_MAX; i++) {
	    if(strstr(nodeName,cb[i].rpcName) != NULL) {
		if(cb[i].funcptr != NULL) {
		    printf("ifx_cwmp_process_soap_body(): about to call func ptr, %s\n",cb[i].rpcName);
		    ret = cb[i].funcptr(childnode, rpcResp);
/*		    if(ret == 1)
			; //Do nothing
		    else if(ret == 0)
			ifx_cwmp_send_rpc_resp(info, rpcResp, id, 0);
		    else if(ret > 1)
			ifx_cwmp_send_rpc_resp(info, (char *)ret, id, 0);
		    else
			goto errorHandler;
*/		}
		else
		    ifx_cwmp_send_fault_response(info, 9000, NULL, 0); 
		break;
	    }
	}

/*    childnode = ixmlNode_getFirstChild(childnode);
    printf("Body3: %s\n", ixmlNode_getNodeName(childnode));
    childnode = ixmlNode_getFirstChild(childnode);
    printf("Body4: %s\n", ixmlNode_getNodeValue(childnode));*/

    return SOAP_SUCCESS;

  errorHandler:
    return SOAP_FAILURE;
}

SOCKINFO * ifx_cwmp_get_sockinfo()
{
    return sockinfo;
}

void ifx_cwmp_set_sockinfo(SOCKINFO *info)
{
    sockinfo = info;
}

/*
 *
 *
 */

int ifx_cwmp_parse_msg(SOCKINFO *info, http_parser_t *parser)
{

    IXML_Document *xmlDoc = NULL;
    IXML_Node *envNode = NULL; 
    IXML_Node *parentnode = NULL, *childnode = NULL;
    IXML_Node *bodynode = NULL;
    //IXML_NodeList *nl = NULL;
    const char *err_str=NULL; 
    char *id=NULL;
    int err;

    err = SOAP_INVALID_ACTION;
    err_str = Soap_Invalid_Action;

    /* Validate Content-Type: text/xml */
/*    if(!has_xml_content_type(&parser->msg)) {
        DEBUGONLY(printf("ifx_cwmp_parse_msg(): Content-Type:text/xml not present in header\n");)
	err = SREQ_BAD_HDR_FORMAT;
	goto error_handler;
    }*/

    /* Parse msg*/
    //printf("entity=%s\n",parser->msg.entity.buf);
    if((err = ixmlParseBufferEx(parser->msg.entity.buf, &xmlDoc)) != IXML_SUCCESS) {
        if( err == IXML_INSUFFICIENT_MEMORY ) {
            err = UPNP_E_OUTOF_MEMORY;
        } else {
            err = SOAP_ACTION_FAILED;
        }
        DEBUGONLY(printf("ifx_cwmp_parse_msg(): xml parse error\n");)
        err_str = "XML error";
        goto error_handler;
    }

    /* Parse through all the nodes till you reach the rpc call node*/
    if((envNode = ixmlNode_getFirstChild((IXML_Node *)xmlDoc)) == NULL) {
	err = ERR_INTERNAL_ERROR;
	err_str = "No SOAP Envelope";
	goto error_handler;
    }

    if(strstr((char *)ixmlNode_getNodeName(envNode), "Envelope") == NULL) {
        DEBUGONLY(printf("ifx_cwmp_parse_msg(): No envelope node\n");)
        printf("ifx_cwmp_parse_msg(): No envelope node\n");
	err = SREQ_BAD_HDR_FORMAT;
	goto error_handler;
    }

    parentnode = ixmlNode_getFirstChild(envNode);
    printf("Parent: %s\n", ixmlNode_getNodeName(parentnode));
    if(strstr((char *)ixmlNode_getNodeName(parentnode),"Header") != NULL) {
	//printf("Header processing\n");
	childnode = ixmlNode_getFirstChild(parentnode);
	//printf("%s, \n", ixmlNode_getNodeName(childnode));
	if(strstr(ixmlNode_getNodeName(childnode), "cwmp:ID") >= 0) {
	    childnode = ixmlNode_getFirstChild(childnode);
	    id = ixmlNode_getNodeValue(childnode);
	    printf("ID=%s\n",id);
	}

	bodynode = ixmlNode_getNextSibling(parentnode);
	if(ifx_cwmp_process_soap_body(info, bodynode, id) != 0)
	    printf("Error processing body\n");

    } else if(strstr((char *)ixmlNode_getNodeName(parentnode),"Body") != NULL) {
	    if(ifx_cwmp_process_soap_body(info, parentnode, id) != 0)
		printf("Error processing body\n");
    } else {
	err = ERR_METHOD_NOT_SUPPORTED;
	goto error_handler;
    }

    err = 0;

  error_handler:
    ixmlDocument_free( xmlDoc );
    if( err != 0 ) {
	printf("err=%d\n", err);
        ifx_cwmp_send_fault_response( info, err, NULL, 0 );
    }
    return 0;
}

/****************************************************************************
*	Function :	soap_device_callback
*
*	Parameters :
*		  IN http_parser_t *parser : Parsed request received by the device
*		  IN http_message_t* request :	HTTP request 
*		  INOUT SOCKINFO *info :	socket info
*
*	Description :	This is a callback called by minisever after receiving 
*		the request from the control point. This function will start 
*		processing the request. It calls handle_invoke_action to handle the
*		SOAP action
*
*	Return :	void
*
*	Note :
****************************************************************************/
void
soap_device_callback( IN http_parser_t * parser,
                      IN http_message_t * request,
                      INOUT SOCKINFO * info )
{
    int err_code;
    const char *err_str;
    memptr action_name;
    IXML_Document *xml_doc = NULL;

    // set default error
    err_code = SOAP_INVALID_ACTION;
    err_str = Soap_Invalid_Action;
 
    //ifx_cwmp_set_sockinfo(info);
    printf("soap_device_callback() called\n");
    if((err_code = ifx_cwmp_parse_msg(info, parser)) == 0)
	return;
	

    // validate: content-type == text/xml
    if( !has_xml_content_type( request ) ) {
        goto error_handler;
    }
    
    // type of request
    if( get_request_type( request, &action_name ) != 0 ) {
	printf("soap: get_request_type failure\n"); // Nirint
        goto error_handler;
    }
    // parse XML
    err_code = ixmlParseBufferEx( request->entity.buf, &xml_doc );
    if( err_code != IXML_SUCCESS ) {
        if( err_code == IXML_INSUFFICIENT_MEMORY ) {
            err_code = UPNP_E_OUTOF_MEMORY;
        } else {
            err_code = SOAP_ACTION_FAILED;
        }

        err_str = "XML error";
        goto error_handler;
    }

    if( action_name.length == 0 ) {
        // query var
        handle_query_variable( info, request, xml_doc );
    } else {
        // invoke action
        handle_invoke_action( info, request, action_name, xml_doc );
    }

    err_code = 0;               // no error

  error_handler:
    ixmlDocument_free( xml_doc );
    if( err_code != 0 ) {
        send_error_response( info, err_code, err_str, request );
    }
}

#endif // EXCLUDE_SOAP

#endif // INCLUDE_DEVICE_APIS
