/*
 * options.c -- DHCP server option packet tools
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "common.h"

#include <dlfcn.h>





/* supported options are easily added here */
const struct dhcp_option dhcp_options[] = {
	/* name[10]	flags					code */
	{"subnet",	OPTION_IP | OPTION_REQ,			0x01},
	{"timezone",	OPTION_S32,				0x02},
	{"router",	OPTION_IP | OPTION_LIST | OPTION_REQ,	0x03},
	{"timesvr",	OPTION_IP | OPTION_LIST,		0x04},
	{"namesvr",	OPTION_IP | OPTION_LIST,		0x05},
	{"dns",		OPTION_IP | OPTION_LIST | OPTION_REQ,	0x06},
	{"logsvr",	OPTION_IP | OPTION_LIST,		0x07},
	{"cookiesvr",	OPTION_IP | OPTION_LIST,		0x08},
	{"lprsvr",	OPTION_IP | OPTION_LIST,		0x09},
	{"hostname",	OPTION_STRING | OPTION_REQ,		0x0c},
	{"bootsize",	OPTION_U16,				0x0d},
	{"domain",	OPTION_STRING | OPTION_REQ,		0x0f},
	{"swapsvr",	OPTION_IP,				0x10},
	{"rootpath",	OPTION_STRING,				0x11},
	{"ipttl",	OPTION_U8,				0x17},
	{"mtu",		OPTION_U16,				0x1a},
	{"broadcast",	OPTION_IP | OPTION_REQ,			0x1c},
	{"ntpsrv",	OPTION_IP | OPTION_LIST,		0x2a},
	{"wins",	OPTION_IP | OPTION_LIST,		0x2c},
	{"requestip",	OPTION_IP,				0x32},
	{"lease",	OPTION_U32,				0x33},
	{"dhcptype",	OPTION_U8,				0x35},
	{"serverid",	OPTION_IP,				0x36},
	{"message",	OPTION_STRING,				0x38},
	{"tftp",	OPTION_STRING,				0x42},
	{"bootfile",	OPTION_STRING,				0x43},
	{"",		0x00,				0x00}
};

/* Lengths of the different option types */
const int option_lengths[] = {
	[OPTION_IP] =		4,
	[OPTION_IP_PAIR] =	8,
	[OPTION_BOOLEAN] =	1,
	[OPTION_STRING] =	1,
	[OPTION_U8] =		1,
	[OPTION_U16] =		2,
	[OPTION_S16] =		2,
	[OPTION_U32] =		4,
	[OPTION_S32] =		4
};


/* get an option with bounds checking (warning, not aligned). */
uint8_t *get_option(struct dhcpMessage *packet, int code)
{
	int i, length;
	uint8_t *optionptr;
	int over = 0, done = 0, curr = OPTION_FIELD;

	optionptr = packet->options;
	i = 0;
	length = 308;
	while (!done) {
		if (i >= length) {
			LOG(LOG_WARNING, "bogus packet, option fields too long.");
			return NULL;
		}
		if (optionptr[i + OPT_CODE] == code) {
			if (i + 1 + optionptr[i + OPT_LEN] >= length) {
				LOG(LOG_WARNING, "bogus packet, option fields too long.");
				return NULL;
			}
			return optionptr + i + 2;
		}
		switch (optionptr[i + OPT_CODE]) {
		case DHCP_PADDING:
			i++;
			break;
		case DHCP_OPTION_OVER:
			if (i + 1 + optionptr[i + OPT_LEN] >= length) {
				LOG(LOG_WARNING, "bogus packet, option fields too long.");
				return NULL;
			}
			over = optionptr[i + 3];
			i += optionptr[OPT_LEN] + 2;
			break;
		case DHCP_END:
			if (curr == OPTION_FIELD && over & FILE_FIELD) {
				optionptr = packet->file;
				i = 0;
				length = 128;
				curr = FILE_FIELD;
			} else if (curr == FILE_FIELD && over & SNAME_FIELD) {
				optionptr = packet->sname;
				i = 0;
				length = 64;
				curr = SNAME_FIELD;
			} else done = 1;
			break;
		default:
			i += optionptr[OPT_LEN + i] + 2;
		}
	}
	return NULL;
}


/* return the position of the 'end' option (no bounds checking) */
int end_option(uint8_t *optionptr)
{
	int i = 0;

	while (optionptr[i] != DHCP_END) {
		if (optionptr[i] == DHCP_PADDING) i++;
		else i += optionptr[i + OPT_LEN] + 2;
	}
	return i;
}


/* add an option string to the options (an option string contains an option code,
 * length, then data) */
int add_option_string(uint8_t *optionptr, uint8_t *string)
{
	int end = end_option(optionptr);

	/* end position + string length + option code/length + end option */
	if (end + string[OPT_LEN] + 2 + 1 >= 308) {
		LOG(LOG_ERR, "Option 0x%02x did not fit into the packet!", string[OPT_CODE]);
		return 0;
	}
	DEBUG(LOG_INFO, "adding option 0x%02x", string[OPT_CODE]);
	memcpy(optionptr + end, string, string[OPT_LEN] + 2);
	optionptr[end + string[OPT_LEN] + 2] = DHCP_END;
	return string[OPT_LEN] + 2;
}


/* add a one to four byte option to a packet */
int add_simple_option(uint8_t *optionptr, uint8_t code, uint32_t data)
{
	char length = 0;
	int i;
	uint8_t option[2 + 4];
	uint8_t *u8;
	uint16_t *u16;
	uint32_t *u32;
	uint32_t aligned;
	u8 = (uint8_t *) &aligned;
	u16 = (uint16_t *) &aligned;
	u32 = &aligned;

	for (i = 0; dhcp_options[i].code; i++)
		if (dhcp_options[i].code == code) {
			length = option_lengths[dhcp_options[i].flags & TYPE_MASK];
		}

	if (!length) {
		DEBUG(LOG_ERR, "Could not add option 0x%02x", code);
		return 0;
	}

	option[OPT_CODE] = code;
	option[OPT_LEN] = length;

	switch (length) {
		case 1: *u8 =  data; break;
		case 2: *u16 = data; break;
		case 4: *u32 = data; break;
	}
	memcpy(option + 2, &aligned, length);
	return add_option_string(optionptr, option);
}


/* find option 'code' in opt_list */
struct option_set *find_option(struct option_set *opt_list, char code)
{
	while (opt_list && opt_list->data[OPT_CODE] < code)
		opt_list = opt_list->next;

	if (opt_list && opt_list->data[OPT_CODE] == code) return opt_list;
	else return NULL;
}


/* add an option to the opt_list */
void attach_option(struct option_set **opt_list, struct dhcp_option *option, char *buffer, int length)
{
	struct option_set *existing, *new, **curr;

	/* add it to an existing option */
	if ((existing = find_option(*opt_list, option->code))) {
		DEBUG(LOG_INFO, "Attaching option %s to existing member of list", option->name);
		if (option->flags & OPTION_LIST) {
			if (existing->data[OPT_LEN] + length <= 255) {
				existing->data = realloc(existing->data,
						existing->data[OPT_LEN] + length + 2);
				memcpy(existing->data + existing->data[OPT_LEN] + 2, buffer, length);
				existing->data[OPT_LEN] += length;
			} /* else, ignore the data, we could put this in a second option in the future */
		} /* else, ignore the new data */
	} else {
		DEBUG(LOG_INFO, "Attaching option %s to list", option->name);

		/* make a new option */
		new = xmalloc(sizeof(struct option_set));
		new->data = xmalloc(length + 2);
		new->data[OPT_CODE] = option->code;
		new->data[OPT_LEN] = length;
		memcpy(new->data + 2, buffer, length);

		curr = opt_list;
		while (*curr && (*curr)->data[OPT_CODE] < option->code)
			curr = &(*curr)->next;

		new->next = *curr;
		*curr = new;
	}
}

#ifdef IFX_CONFIG_TR111_DEVICE_ASSOCIATION 
/*

This function takes pointer to the first enterprise no of the vendor specific options

    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  option-code  |  option-len   |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |      enterprise-number1       |
   |                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |   data-len1   |               |
   +-+-+-+-+-+-+-+-+ option-data1  |
   /                               /
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ ----
   |      enterprise-number2       |   ^
   |                               |   |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+   |
   |   data-len2   |               | optional
   +-+-+-+-+-+-+-+-+ option-data2  |   |
   /                               /   |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+   |
   ~            ...                ~   V
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ ----

And then returns the pointer to Vendor specific info for the specified Enterprise number.
If the info relatedd to enterprise no is not present, then NULL is returned.


pVSO : ponter to the first enterprise no field below the VSI.
iEnterpriseNo : is the enterprise no of interest
option_len : is the length of the buffer pointer to by pVSO as mentioned in the option-len of VSI.

*/

uint8_t *get_vsi_option(uint8_t *pVSO, uint32_t iEnterpriseNo, uint8_t option_len)
{
	int i, length;
	uint8_t *optionptr=NULL;
	uint32_t *pEntNo = NULL;
	int done=0;

	return pVSO;

	DEBUG(LOG_ERR, " GET_VSI_OPTION  called\n");
	optionptr = pVSO;
	i = 0;
	length = option_len;
	while (!done) {
		if (i >= length) {
			LOG(LOG_WARNING, "bogus packet, option fields too long.");
			return NULL;
		}
		
		pEntNo = (uint32_t *)(optionptr +i);

		if (*pEntNo == iEnterpriseNo) {
			if (i + OPT_VSI_ENTPNO_LEN + optionptr[i + OPT_VSI_ENTPNO_LEN] >= length) {
				LOG(LOG_WARNING, "bogus packet, option fields too long.");
				return NULL;
			}
			return optionptr + i + OPT_VSI_ENTPNO_LEN;
		}else{
			i += OPT_VSI_ENTPNO_LEN + optionptr[i + OPT_VSI_ENTPNO_LEN];

		}
	}
	DEBUG(LOG_ERR, " GET_VSI_OPTION  over\n");
	return NULL;
}


/******************************************************************
This function posts a message with appropriats options and 
null string when the entry for Gateway or Device is to be removed
Function used by :
DEVICE to write empty string for gateway, ipaddr = 0
GATEWAY to delete the ipaddr of a device whose lease has expired
*******************************************************************/
int writenullmsg(uint32_t iIpaddr){
	int iRetrn = 0;

	iRetrn = writemsg( TR69_FIFO, NULL, 2,iIpaddr);	

	return iRetrn;
}

/* 
This function takes pointer to Vendor Specific info below the identified enterprise no.

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |   data-len1   |               |
   +-+-+-+-+-+-+-+-+ option-data   |
   /                               /
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ 

Where option-data is a collection of sub-options each with the format 

   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  subopt-code  |  subopt-len   |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   /        sub-option-data        /
   /                               /
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

And fills the string in the form 
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
   / sub-option-data1\0suboption-data2\0suboption-data3\0 /
   /                               						  /
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-++-+-+-+-+-+-+-+-+-+-+-+


RETURN VALUES :
	SUCCESS 	string length for output
	FAILURE		-1

*/


int form_vsi_string(uint8_t *pVSD,char *string)
{
        int i;
        uint8_t *optionptr=NULL;
        uint8_t data_len, option_len;
	uint8_t string_len = 0;

        optionptr = pVSD;
	data_len = optionptr[0];
	optionptr++;
        i = 0;

		DEBUG(LOG_ERR, " FORM_VSI_STRING called\n");
        while ( i < data_len) {
#ifdef IFX_CONFIG_TR69_DEVICE
			if(( optionptr[i + SUBOPT_CODE] < 0x04) || ( optionptr[i + SUBOPT_CODE] > 0x06)){
				LOG(LOG_WARNING, "bogus packet, invalid sub-option code from device");	
				return -1; //failure
			} 
#else
			if(( optionptr[i + SUBOPT_CODE] < 0x01) || ( optionptr[i + SUBOPT_CODE] > 0x03)){
				LOG(LOG_WARNING, "bogus packet, invalid sub-option code from device");	
				return -1; //failure
			}
#endif

			option_len = optionptr[i + SUBOPT_LEN];

			if( i + 1 +  option_len > data_len) {
					LOG(LOG_WARNING, "bogus packet, option fields too long.");
					return -1 ; // failure
			}

			DEBUG(LOG_ERR, " FORM_VSI_STRING : copying option to be sent\n");
			memcpy(string+ string_len,optionptr + i + SUBOPT_DATA, option_len);
				
			i += 1 + 1 + option_len; // move the offset to next VSD
			string_len = string_len + option_len;	// update the total string size
			*(string + string_len)= '\0'; //add a null in the end of thie string
			string_len++;
		}
			

		DEBUG(LOG_ERR, " FORM_VSI_STRING done\n");
		return string_len-1; //success
}






/************************************************************
This function writes a message to file.
The structure of the message is 

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |   data-len    |               |
   +-+-+-+-+-+-+-+-+      data     |
   /                               /
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ 


************************************************************/

/********************************************************************************************
The operation defined are :
 1	Add	Add the information in the table (requires add object and write of parameters
 2	Del 	GATEWAY :Delete the information from the table (required delete of object) 
		DEVICE : Clear GatewayInfo

The operation to be done is also reflected in the Reserved field of the message
The data of the message sent is only needed in ADD
For a delete or write only the ipaddr matters 
********************************************************************************************/
int writemsg(char *file, uint8_t *msg, uint8_t operation, uint32_t ipaddr){
	int iFd, iRet;
	x_IFX_IPC_Msg xMsg = {0};
	char string[380] = {'\0'};
	uint16_t string_len = 0;

	DEBUG(LOG_ERR, " WRITEMSG CALLED\n");
	memset(&xMsg, 0, sizeof(x_IFX_IPC_Msg));

	if((operation == 1)&&(msg == NULL))
		return -1; //error

	
	xMsg.xHdr.ucFrom = IFX_IPC_APP_ID_DHCP;
	xMsg.xHdr.ucTo = operation;
	xMsg.xHdr.uiReserved = htonl(ipaddr);
	
	// data is meaningful only in add
	if(operation == 1){
		DEBUG(LOG_ERR, " WRITEMSG : FORM STRING TO BE POSTED \n");
		string_len =  form_vsi_string(msg+4, string);
		if(string_len < 0){
			DEBUG(LOG_ERR, "Failure in forminmg string data");
			return -1;
		}

		xMsg.xHdr.unMsgSize =  htons(string_len);  
		memcpy(xMsg.acMsg, (void *)string, string_len);
	}

	iFd = open(file,O_WRONLY | O_NONBLOCK);
	DEBUG(LOG_ERR, " WRITEMSG : FIFO OPENED WITH FD : %d \n", iFd);
	if(iFd > -1 ){  	//now post this msg to TR69 stack
		iRet = write( iFd,(void *)&xMsg, string_len + IFX_IPC_HDR_SIZE);
		close(iFd);
		if(iRet <=0){
			DEBUG(LOG_ERR, "Failure in writing data");
			return -1;
		}
		DEBUG(LOG_ERR, " WRITEMSG DONE \n");
	}
	return iRet;
}

/* This function initializes the variables to be used in TR-111 option 125
*/
char subopt1[IFX_DEVM_DHCP_PARAM_SIZE];
char subopt2[IFX_DEVM_DHCP_PARAM_SIZE];
char subopt3[IFX_DEVM_DHCP_PARAM_SIZE];
char subopt4[IFX_DEVM_DHCP_PARAM_SIZE];
char subopt5[IFX_DEVM_DHCP_PARAM_SIZE];
char subopt6[IFX_DEVM_DHCP_PARAM_SIZE];

void ifx_init_tr111_vars(){
	int iRet = 0;
	char sValue[IFX_DEVM_DHCP_PARAM_SIZE] = {0};
	void *dlHandle = NULL;
	int (*func_ptr)(const char *, const char *, char *, char *) = NULL;
	char *error = NULL;
	char sVal[MAX_FILELINE_LEN]={0};
	int iFD=-1;
    	char caBuff[32] = { 0 };
    	char *psTmp = NULL;
	

	dlHandle = dlopen("libifx_common.so",RTLD_LAZY);
	if(dlHandle) {
		*(void **) &func_ptr = dlsym(dlHandle,"ifx_GetCfgData");
		if((error = dlerror()) != NULL) {
			DEBUG(LOG_ERR, "could not find function ifx_GetCfgData in libifx_common.so\n");
			DEBUG(LOG_ERR, "dlopen failed !! taking backup values \n");
			goto BACKUP;
		}

		if ( (int)(* func_ptr)(IFX_SYS_CONF, IFX_DEVICE_INFO_SECTION, "device_info_oui", sValue) == 1){

			// frame the OUI from the MAC address - logic picked from IFXAPIs/src/mapi/ifx_api_dev_info.c:130
			memset(sVal, 0x00, sizeof(sVal));
			system("upgrade mac_get 2 >/tmp/macget");
			iFD = open("/tmp/macget", O_RDONLY);
			if(iFD < 0)
			{
				IFX_DBG("failed to open /tmp/macget\n");
				return -1;
			}
			read(iFD, caBuff, sizeof(caBuff));
			psTmp = strtok(caBuff, ":");
			if(psTmp == NULL)
			{
				IFX_DBG("Unable to parse version /tmp/macget\n");
				return -1;
			}
			strcpy(sVal, psTmp);
			psTmp = strtok(NULL, ":");
			strcat(sVal, psTmp);
			psTmp = strtok(NULL, ":");
			strcat(sVal, psTmp);
			close(iFD);
			system("rm /tmp/macget");

			if ((strlen(sVal) + 1) > IFX_DEVM_DHCP_PARAM_SIZE)
			{
				goto BACKUP;
			}
			strcpy(sValue, sVal);

#ifdef IFX_CONFIG_TR69_DEVICE 	//  it is a device
			strcpy(subopt1,sValue);
#else	// it is a gateway
			strcpy(subopt4,sValue);
#endif
		}else{
			goto BACKUP;
		}

		if ( (int)(* func_ptr)(IFX_SYS_CONF, IFX_DEVICE_INFO_SECTION, "device_info_sernum", sValue) == 1){
#ifdef IFX_CONFIG_TR69_DEVICE 	//  it is a device
			strcpy(subopt2,sValue);
#else	// it is a gateway
			strcpy(subopt5,sValue);
#endif
		}else{
			goto BACKUP;
		}

		if ( (int)(* func_ptr)(IFX_SYS_CONF, IFX_DEVICE_INFO_SECTION, "device_info_prodclass", sValue) == 1){
#ifdef IFX_CONFIG_TR69_DEVICE 	//  it is a device
			strcpy(subopt3,sValue);
#else	// it is a gateway
			strcpy(subopt6,sValue);
#endif
		}else{
			goto BACKUP;
		}

	} else {
		DEBUG(LOG_ERR, "dlopen failed !! taking backup values \n");
BACKUP : 
#ifdef IFX_CONFIG_TR69_DEVICE 	//  it is a device
		strcpy(subopt1,"001212");
		strcpy(subopt2,"DMA67");
		strcpy(subopt3,"DEVICE");
#else	// it is a gateway
		strcpy(subopt4,"011111");
		strcpy(subopt5,"DMA66");
		strcpy(subopt6,"GATEWAY");
#endif

}

	dlclose(dlHandle);
	return;
}

#endif
